﻿using System;
using System.IO;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.Numerics;
using NAudio.Wave;
using MathNet.Numerics;
using MathNet.Numerics.IntegralTransforms;


namespace Speech2Music
{
    public partial class Form1 : Form
    {
        #region class members
        string musicWAVfile = null;                 //from user input
        string musicWAVfileMono = "musicMono.wav";  //used after mono transform
        string speechWAVfile = null;                //from user input
        string speechWAVfileMono = "speechMono.wav";//used after mono transform
        string shiftWAVfile = null;               //where to save the shift file 
        byte[] speechheader = null;                 //header of the speech file
        byte[] speechdata = null;                   //data of the speech file
        byte[] shiftdata = null;                    //data of the shift file
        byte[] copydata = null;                     //data that will alter - copy of the speech data
        byte[] header = null;                       //used for saving the wav file
        byte[] data = null;                         //used for saving the wav file
        float[] data_l = null;                      //speech left channel
        float[] data_r = null;                      //speech right channel
        int sampleRate = 44100;                     //standard sample rate
        int bitsPerSample = 16;                     //standard
        int fftSize;                        //standard -- can be altered

        //AudioFileReader -- used for reading the wav files
        float[] musicsamples = null;
        float[] speechsamples = null;

        //From Capstone 2015
        private WaveFileReader waveReader = null;
        private DirectSoundOut waveOut = null;
        //for timers/tracking wave position
        public Timer playTimer;
        double wavPosition;
        double lastWavPosition;
        //Second form
        About AboutForm;

        // for saving the shift file
        // source: url: https://msdn.microsoft.com/en-us/library/system.windows.forms.folderbrowserdialog.aspx
        
        #endregion

        public Form1()
        {
            WindowState = FormWindowState.Maximized; // makes form full screen
            
            InitializeComponent();
        }

        #region Menu Bar Methods
        //Open the speech file
        private void openSpeechWavToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog open = new OpenFileDialog();
            open.Filter = "Wave File (*,wav)|*.wav;"; //make sure it if of wav type
            if (open.ShowDialog() != DialogResult.OK) return;
            speechWAVfile = open.FileName;
            //open in CustomWaveViewer
            speechWaveViewer.WaveStream = new NAudio.Wave.WaveFileReader(speechWAVfile);
        }

        //Open the music file
        private void openMusicWavToolStripMenuItem_Click(object sender, EventArgs e)
        {
            OpenFileDialog open = new OpenFileDialog();
            open.Filter = "Wave File (*,wav)|*.wav;"; //make sure it is of wav file type
            if (open.ShowDialog() != DialogResult.OK) return;
            musicWAVfile = open.FileName;
            musicWaveViewer.WaveStream = new NAudio.Wave.WaveFileReader(musicWAVfile);
        }

        //Open the About form
        private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
        {
            int count = 0;
            FormCollection fc = Application.OpenForms;

            foreach (Form frm in fc)
                count++;
            if (count == 1) //only the main form is open
            {
                // Create a new instance of the Form2 class
                AboutForm = new About();
                // Show the settings form
                AboutForm.Show();
            }
            //bring attention to the form already opens
            else
                AboutForm.Activate();
        }

        #endregion


        #region Button Click handlers and player mechanics
        //Play the speech wav
        private void speechPlaybtn_Click(object sender, EventArgs e)
        {
            if (speechWAVfile != null)
            {
                playWAV(speechWAVfile);
                speechPlayLine();
            }
        }

        //Play the music wav
        private void musicPlaybtn_Click(object sender, EventArgs e)
        {
            if (musicWAVfile != null)
            {
                playWAV(musicWAVfile);
                
            }
        }

        //Play the shift wav
        private void shiftPlaybtn_Click(object sender, EventArgs e)
        {
            if (shiftWAVfile != null)
            {
                playWAV(shiftWAVfile);
                
            }
        }

        //Stop the speech wav
        private void speechStopbtn_Click(object sender, EventArgs e)
        {
            stopWAV(speechWAVfile);
        }

        //Stop the music wav
        private void musicStopbtn_Click(object sender, EventArgs e)
        {
            stopWAV(musicWAVfile);
        }

        //Stop the shift wav
        private void shiftStopbtn_Click(object sender, EventArgs e)
        {
            stopWAV(shiftWAVfile);
        }

        //universal function for playing the data of any wave viewer
        private void playWAV(string waveFile)
        {
            disposeWave();
            waveReader = new WaveFileReader(waveFile);
            waveOut = new DirectSoundOut();
            waveOut.Init(new WaveChannel32(waveReader));

            if (waveOut != null)
            {
                waveOut.Play();
            }
        }

        private void pauseWAV(string waveFile)
        {
            if (waveOut != null)
            {
                waveOut.Pause();
            }
        }

        private void stopWAV(string waveFile)
        {
            if (waveOut != null)
            {
                waveOut.Stop();
            }

        }

        //PLAYING TIMER / POSITION TRACKER -- FROM CAPSTONE 2015
        //SEPERATE FUNCTIONS FOR EACH WAVEVIEWER

        //PLAY LINE FNCTS FOR THE SPEECH WAVE VIEWER
        public void speechPlayLine()
        {
            if (playTimer != null)
                disposePlayTimer();
            wavPosition = (double)(speechWaveViewer.pixelPerMillisecond() * waveReader.CurrentTime.TotalMilliseconds);
            lastWavPosition = wavPosition;
            speechWaveViewer.drawWavePosition(wavPosition);

            playTimer = new Timer();
            playTimer.Interval = 1;
            playTimer.Tick += new EventHandler(speechPlayTimer_Tick);
            playTimer.Enabled = true;
            playTimer.Start();
        }
        private void speechPlayTimer_Tick(object sender, EventArgs e)
        {
            wavPosition = (double)(speechWaveViewer.pixelPerMillisecond() * waveReader.CurrentTime.TotalMilliseconds);
            speechWaveViewer.drawWavePosition(lastWavPosition);
            speechWaveViewer.drawWavePosition(wavPosition);
            lastWavPosition = wavPosition;

            if (wavPosition >= speechWaveViewer.Width)
            {
                playTimer.Enabled = false;
                speechWaveViewer.drawWavePosition(lastWavPosition);
                disposePlayTimer();
            }

        }

        //PLAY LINE FNCTS FOR THE MUSIC WAVE VIEWER
        public void musicPlayLine()
        {
            if (playTimer != null)
                disposePlayTimer();
            wavPosition = (double)(musicWaveViewer.pixelPerMillisecond() * waveReader.CurrentTime.TotalMilliseconds);
            lastWavPosition = wavPosition;
            musicWaveViewer.drawWavePosition(wavPosition);

            playTimer = new Timer();
            playTimer.Interval = 1;
            playTimer.Tick += new EventHandler(musicPlayTimer_Tick);
            playTimer.Enabled = true;
            playTimer.Start();
        }
        private void musicPlayTimer_Tick(object sender, EventArgs e)
        {
            wavPosition = (double)(musicWaveViewer.pixelPerMillisecond() * waveReader.CurrentTime.TotalMilliseconds);
            musicWaveViewer.drawWavePosition(lastWavPosition);
            musicWaveViewer.drawWavePosition(wavPosition);
            lastWavPosition = wavPosition;

            if (wavPosition >= musicWaveViewer.Width)
            {
                playTimer.Enabled = false;
                musicWaveViewer.drawWavePosition(lastWavPosition);
                disposePlayTimer();
            }

        }

        private void disposePlayTimer()
        {
            if (playTimer != null)
            {
                playTimer.Stop();
                playTimer.Enabled = false;
                playTimer.Dispose();
                playTimer = null;
            }

        }


        //Clean up -- from 2015 Capstone - WORKS
        private void disposeWave()
        {
            if (waveOut != null)
            {
                if (waveOut.PlaybackState == NAudio.Wave.PlaybackState.Playing)
                    waveOut.Stop();
                waveOut.Dispose();
                waveOut = null;
            }

            if (waveReader != null)
            {
                waveReader.Dispose();
                waveReader = null;
            }
        }

        //Saves the shift wav file -- want the user to be able to pick where it goes
        //still needs work I think
        private void shiftSavebtn_Click(object sender, EventArgs e)
        {
            SaveFileDialog sfd = new SaveFileDialog();
            sfd.Filter = "Wave File (*,wav)|*.wav;";
            sfd.Title = "Save Wav File As";
            sfd.ShowDialog();

            if (sfd.FileName != "")
            {
                byte[] bytes = File.ReadAllBytes(shiftWAVfile);
                File.WriteAllBytes(sfd.FileName, bytes);
            }
            
        }
        #endregion


        //function to shift the frequency of the speech to match the frequency of the music
        private void shiftbtn_Click(object sender, EventArgs e)
        {
            //make sure the user wants to continue
            if (shiftWAVfile != null)
            {
                var userConfirm = MessageBox.Show("If you continue, any unsaved data will be lost.", "Continue?", MessageBoxButtons.YesNo);
                if (userConfirm == DialogResult.No)
                    return;
                //still here close the wave stream
                shiftWaveViewer.WaveStream.Close();
            }

            disposeWave();

            if (musicWAVfile != null && speechWAVfile != null)
            {
                processinglbl.Text = "PROCESSING";
                //make wave file mono
                StereoToMono(musicWAVfile, musicWAVfileMono);
                StereoToMono(speechWAVfile, speechWAVfileMono);

                AudioFileReader mAudio = new AudioFileReader(musicWAVfileMono);
                AudioFileReader sAudio = new AudioFileReader(speechWAVfileMono);

                //Isampleprovider
                //AudioFileReader reader = new AudioFileReader(musicWAVfileMono);
                //MySampleProvider sp = new MySampleProvider(reader);


                byte[] musicwav = File.ReadAllBytes(musicWAVfileMono); //get all of the bytes
                byte[] speechwav = File.ReadAllBytes(speechWAVfileMono);
                shiftdata = new byte[speechwav.Length];
                int totalLength = musicwav.Length;
                //start loop -- Will have to read chunks in AudioFileReader.Read
                if (totalLength == speechwav.Length)
                {
                    int count = 0;
                    //int lengthChunks = speechwav.Length / numChunks;
                    int index = 0;
                    musicsamples = new float[musicwav.Length / 2];
                    speechsamples = new float[speechwav.Length / 2];

                    if (onerb.Checked) { fftSize = 44100; }
                    else { fftSize = speechsamples.Length; } //NOTE: this is kind of cheating -- only works if clip is 5 seconds
                    
                    double divisor = (double)fftSize / (double)sampleRate;

                    int numChunks = musicsamples.Length / fftSize;
                    if (numChunks == 0) { count = 1; } //only enough for one loop
                    //where the ending frequencies will be listed
                    float[] speechTotalF = new float[numChunks];
                    float[] musicTotalF = new float[numChunks];

                    //get the wave data of the speech file -- the one to be altered
                    GetWaveData(speechWAVfileMono, out speechheader, out speechdata, out sampleRate, out data_l, out data_r);
                    shiftdata = new byte[speechdata.Length];
                    copydata = new byte[speechdata.Length];
                    Array.Copy(speechdata, shiftdata, speechdata.Length);
                    Array.Copy(speechdata, copydata, speechdata.Length);

                    while (count < numChunks)
                    {
                        //read the chunk --audio file reader
                        mAudio.Read(musicsamples, 0, musicsamples.Length);
                        sAudio.Read(speechsamples, 0, speechsamples.Length);

                        //divide the samples into workable parts
                        float[] musicchunk = new float[fftSize];
                        int j = 0;
                        for (int i = index; i < index + fftSize; i++)
                        {
                            musicchunk[j] = musicsamples[i];
                            j++;
                        }
                        //and again for the speech
                        float[] speechchunk = new float[fftSize];
                        j = 0;
                        for (int i = index; i < index + fftSize; i++)
                        {
                            speechchunk[j] = speechsamples[i];
                            j++;
                        }

                        //Run FFt on the samples
                        float speechfreq = FFTWavFile(speechchunk);
                        speechfreq = speechfreq / (float)divisor;
                        //speechTotalF[count] = speechfreq / (float)divisor;
                        float musicfreq = FFTWavFile(musicchunk);
                        musicfreq = musicfreq / (float)divisor;
                        //musicTotalF[count] = musicfreq / (float)divisor;

                        //NOW SHIFT!!!!!!!
                        ShiftValue(speechfreq, musicfreq, fftSize, count);

                        //update index
                        index += fftSize;
                        count++;
                    }
                    //write the final shift
                    WriteShift();
                    //close audio readers
                    mAudio.Close();
                    sAudio.Close();

                }
                processinglbl.Text = "";
            }
        }

        public void StereoToMono(string sourceFile, string outputFile)
        {
            using (WaveFileReader waveFileReader = new WaveFileReader(sourceFile))
            {
                sampleRate = waveFileReader.WaveFormat.SampleRate;
                bitsPerSample = waveFileReader.WaveFormat.BitsPerSample;
                var outFormat = new WaveFormat(waveFileReader.WaveFormat.SampleRate, 1);
                using (var resampler = new MediaFoundationResampler(waveFileReader, outFormat))
                {
                    if (File.Exists(outputFile))
                        File.Delete(outputFile);
                    WaveFileWriter.CreateWaveFile(outputFile, resampler);
                }
            }
        }

        public void WriteShift()
        {
            shiftWAVfile = "TestPitchShift.wav";
            if (File.Exists(shiftWAVfile))
                File.Delete(shiftWAVfile);
            
            using (BinaryWriter writer = new BinaryWriter(File.OpenWrite(shiftWAVfile)))
            {
                writer.Write(speechheader);
                writer.Write(shiftdata);
            }
            //open the shift wave in the wave viewer

            shiftWaveViewer.WaveStream = new NAudio.Wave.WaveFileReader(shiftWAVfile);
            
        }

        #region Methods for FFT and frequency detection
        public static float FFTWavFile(float[] samples)
        {

            //change float samples to Complex values
            Complex[] cdata = new Complex[samples.Length];
            for (int i = 0; i < cdata.Length; i++)
            {
                Complex temp = new Complex(samples[i], 0);
                cdata[i] = temp;
            }

            Fourier.Forward(cdata); //run the MathNet FFT on complex input

            //File.WriteAllLines(@"C:\Users\Chance\OneDrive\Capstone\TestProjects\MathNetFFT.txt", cdata.Select(d => d.ToString()));

            double largest = 0;
            double[] reals = new double[cdata.Length / 2]; //should only need the first half of the array
            for (int i = 0; i < reals.Length; i++)
            {
                reals[i] = cdata[i].Real;
            }

            largest = reals.Max();
            int index = reals.ToList().IndexOf(largest); //index is what we need
            float freq = index;//index is 10 times larger than what it should be not sure why

            return freq;
        }

        //sf is the frequency of the speech and mf is the frequency of the music
        public int ShiftOctave(float sf, float mf)
        {
            //how many octives
            int octaves = 0;
            int total = octaves;
            float copyFreq = mf;
            //incease octaves loop
            while (sf > (copyFreq * 2))
            {
                octaves--;
                copyFreq = copyFreq * 2;
            }
            //decrease octaves loop
            while (copyFreq > (sf * 2))
            {
                octaves++;
                copyFreq = copyFreq / 2;
            }
            total = octaves;
            //shift for octaves
            while (octaves > 0)
            {
                //pitchshft by 2f
                // Apply Pitch Shifting
                if (data_l != null)
                    PitchShifter.PitchShift(2f, data_l.Length, (long)1024, (long)10, sampleRate, data_l);

                ResetChannels(data_l, data_r, ref shiftdata); //places left and right chanels back & altered

                octaves--;
            }

            while (octaves < 0)
            {
                //pitchshift by 0.5f
                // Apply Pitch Shifting
                if (data_l != null)
                    PitchShifter.PitchShift(0.5f, data_l.Length, (long)1024, (long)10, sampleRate, data_l);

                ResetChannels(data_l, data_r, ref shiftdata); //places left and right chanels back & altered

                octaves++;
            }
            return total;
        }

        //sf is the frequency of the speech and mf is the frequency of the music
        public void ShiftValue(float sf, float mf, int fftSize, int count)
        {
            float difference = mf / sf; //the difference in the frequency
            int length = fftSize * 2;// times 2 because of size difference between float and byte

            // start with octave shifts
            int numshifts = ShiftOctave(sf, mf); //number of octave shifts
            //adjust difference for octave shifts
            //only one of the while loops will run
            while (numshifts > 0) //shifts went up in octaves
            {
                mf = mf / 2;
                difference = mf / sf;
                numshifts--;
            }
            while (numshifts < 0) //shifts went down in octaves
            {
                mf = mf * 2;
                difference = mf / sf;
                numshifts++;
            }
            //shift by the remaining difference
            // 0.5 < difference < 2.0 --must be this way for the pitch shift
            if (data_l != null)
                PitchShifter.PitchShift(difference, data_l.Length, (long)1024, (long)10, sampleRate, data_l);

            ResetChannels(data_l, data_r, ref copydata); //places left and right chanels back & altered

            //so we do not go over the bounds of the source
            if ((length * count) + length > copydata.Length)
            {
                length = copydata.Length - (length * count);
            }

            Array.Copy(copydata, length * count, shiftdata, length * count, length); // times 2 because of size difference between float and byte

            //save shift data
            shiftWAVfile = "TestPitchShift.wav";
            if (File.Exists(shiftWAVfile))
                File.Delete(shiftWAVfile);
            

            using (BinaryWriter writer = new BinaryWriter(File.OpenWrite(shiftWAVfile)))
            {
                writer.Write(speechheader);
                writer.Write(shiftdata);
                //writer.Close();
            }
        }
        #endregion

        #region Methods for PitchShifting
        //FUNCTIONS FOR PITCH SHIFTING
        // Returns left and right float arrays. 'right' will be null if sound is mono.
        public static void GetWaveData(string filename, out byte[] header, out byte[] data, out int sampleRate, out float[] left, out float[] right)
        {
            byte[] wav = File.ReadAllBytes(filename);

            // Determine if mono or stereo
            int channels = wav[22];     // Forget byte 23 as 99.999% of WAVs are 1 or 2 channels

            // Get sample rate 
            //Returns a 32-bit signed integer converted from four bytes at a specified position in a byte array.
            //the sample rate of the wav file         
            sampleRate = BitConverter.ToInt32(wav, 24); //Returns a 32-bit signed integer converted from four bytes at a specified position in a byte array.

            int pos = 12;

            // Keep iterating until we find the data chunk (i.e. 64 61 74 61 ...... (i.e. 100 97 116 97 in decimal))
            while (!(wav[pos] == 100 && wav[pos + 1] == 97 && wav[pos + 2] == 116 && wav[pos + 3] == 97))
            {
                pos += 4;
                int chunkSize = wav[pos] + wav[pos + 1] * 256 + wav[pos + 2] * 65536 + wav[pos + 3] * 16777216;
                pos += 4 + chunkSize;
            }

            pos += 4;

            int subchunk2Size = BitConverter.ToInt32(wav, pos);
            pos += 4;

            // Pos is now positioned to start of actual sound data.
            int samples = subchunk2Size / 2;     // 2 bytes per sample (16 bit sound mono)
            if (channels == 2)
                samples /= 2;        // 4 bytes per sample (16 bit stereo)

            // Allocate memory (right will be null if only mono sound)
            left = new float[samples];

            if (channels == 2)
                right = new float[samples];
            else
                right = null;

            header = new byte[pos];
            Array.Copy(wav, header, pos);

            data = new byte[subchunk2Size];
            Array.Copy(wav, pos, data, 0, subchunk2Size);

            // Write to float array/s:
            int i = 0;
            while (pos < subchunk2Size)
            {

                left[i] = BytesToNormalized_16(wav[pos], wav[pos + 1]);
                pos += 2;
                if (channels == 2)
                {
                    right[i] = BytesToNormalized_16(wav[pos], wav[pos + 1]);
                    pos += 2;
                }
                i++;
            }
        }

        // Return byte data from left and right float data. Ignore right when sound is mono
        public static void ResetChannels(float[] left, float[] right, ref byte[] data)
        {
            // Calculate k
            // This value will be used to convert float to Int16
            // We are not using Int16.Max to avoid peaks due to overflow conversions            
            float k = (float)Int16.MaxValue / left.Select(x => Math.Abs(x)).Max();

            // Revert data to byte format
            Array.Clear(data, 0, data.Length);
            int dataLenght = left.Length;
            int byteId = -1;
            using (BinaryWriter writer = new BinaryWriter(new MemoryStream(data)))
            {
                for (int i = 0; i < dataLenght; i++)
                {
                    byte byte1 = 0;
                    byte byte2 = 0;

                    byteId++;
                    NormalizedToBytes_16(left[i], k, out byte1, out byte2);
                    writer.Write(byte1);
                    writer.Write(byte2);

                    if (right != null)
                    {
                        byteId++;
                        NormalizedToBytes_16(right[i], k, out byte1, out byte2);
                        writer.Write(byte1);
                        writer.Write(byte2);
                    }
                }
            }
        }

        // Convert two bytes to one double in the range -1 to 1
        static float BytesToNormalized_16(byte firstByte, byte secondByte)
        {
            // convert two bytes to one short (little endian)
            short s = (short)((secondByte << 8) | firstByte);
            // convert to range from -1 to (just below) 1
            return s / 32678f;
        }

        // Convert a float value into two bytes (use k as conversion value and not Int16.MaxValue to avoid peaks)
        static void NormalizedToBytes_16(float value, float k, out byte firstByte, out byte secondByte)
        {
            short s = (short)(value * k);
            firstByte = (byte)(s & 0x00FF);
            secondByte = (byte)(s >> 8);
        }

        #endregion

    }
}
